@@ -2,10 +2,11 @@
 # Copyright (c) 2008-2010 Erik Svensson <erik.public@gmail.com>
 # Licensed under the MIT license.
 
-from constants import *
-from transmission import TransmissionError, Torrent, Session, Client
+from transmissionrpc.constants import *
+from transmissionrpc.httphandler import HTTPHandlerError, HTTPHandler
+from transmissionrpc.transmission import TransmissionError, Torrent, Session, Client, DefaultHTTPHandler
 
 __author__    = u'Erik Svensson <erik.public@gmail.com>'
-__version__   = u'0.4'
+__version__   = u'0.5'
 __copyright__ = u'Copyright (c) 2008-2010 Erik Svensson'
 __license__   = u'MIT'
