@@ -2,29 +2,29 @@
 # Copyright (c) 2008-2010 Erik Svensson <erik.public@gmail.com>
 # Licensed under the MIT license.
 
-import sys, os, time, datetime
+import sys, os, re, time, datetime
 import warnings
-import re, urlparse
-import httplib, urllib2, base64, socket
+import httplib, urllib2, urlparse, base64
 
 try:
     import json
 except ImportError:
     import simplejson as json
 
-from constants import *
-from utils import *
+from transmissionrpc.constants import DEFAULT_PORT, DEFAULT_TIMEOUT, STATUS, PRIORITY
+from transmissionrpc.utils import logger, format_timedelta, get_arguments, make_rpc_name, argument_value_convert, rpc_bool
+from transmissionrpc.httphandler import HTTPHandlerError, HTTPHandler
 
 class TransmissionError(Exception):
     def __init__(self, message='', original=None):
-        #Exception.__init__(self, message)
+        Exception.__init__(self)
         self.message = message
         self.original = original
 
     def __str__(self):
         if self.original:
             original_name = type(self.original).__name__
-            return '%s Original exception: %s, "%s"' % (self.message, original_name, self.original.args)
+            return '%s Original exception: %s, "%s"' % (self.message, original_name, str(self.original))
         else:
             return self.message
 
@@ -69,7 +69,6 @@ class Torrent(object):
             files = self.fields['files']
             priorities = self.fields['priorities']
             wanted = self.fields['wanted']
-            index = 1
             for item in zip(indicies, files, priorities, wanted):
                 selected = True if item[3] else False
                 priority = PRIORITY[item[2]]
@@ -84,7 +83,7 @@ class Torrent(object):
     def __getattr__(self, name):
         try:
             return self.fields[name]
-        except KeyError, e:
+        except KeyError:
             raise AttributeError('No attribute %s' % name)
 
     @property
@@ -146,6 +145,11 @@ class Torrent(object):
             return 'unknown'
         else:
             return format_timedelta(self.eta)
+    
+    @property
+    def priority(self):
+        """Get the priority as string."""
+        return PRIORITY[self.fields['bandwidthPriority']]
 
 class Session(object):
     """
@@ -157,9 +161,10 @@ class Session(object):
     ``download-dir`` -> ``download_dir``.
     """
 
-    def __init__(self, fields={}):
+    def __init__(self, fields=None):
         self.fields = {}
-        self.update(fields)
+        if fields != None:
+            self.update(fields)
 
     def update(self, other):
         """Update the session data from a session arguments dictinary"""
@@ -178,7 +183,7 @@ class Session(object):
     def __getattr__(self, name):
         try:
             return self.fields[name]
-        except KeyError, e:
+        except KeyError:
             raise AttributeError('No attribute %s' % name)
 
     def __str__(self):
@@ -187,107 +192,154 @@ class Session(object):
             text += "% 32s: %s\n" % (k[-32:], self.fields[k])
         return text
 
+class DefaultHTTPHandler(HTTPHandler):
+    def __init__(self):
+        HTTPHandler.__init__(self)
+    
+    def set_authentication(self, uri, login, password):    
+        password_manager = urllib2.HTTPPasswordMgrWithDefaultRealm()
+        password_manager.add_password(realm=None, uri=uri, user=login, passwd=password)
+        opener = urllib2.build_opener(
+            urllib2.HTTPBasicAuthHandler(password_manager)
+            , urllib2.HTTPDigestAuthHandler(password_manager)
+            )
+        urllib2.install_opener(opener)
+    
+    def request(self, url, query, headers, timeout):        
+        request = urllib2.Request(url, query, headers)
+        try:
+            if (sys.version_info[0] == 2 and sys.version_info[1] > 5) or sys.version_info[0] > 2:
+                logger.debug('Using timeout %f s' % (timeout))
+                response = urllib2.urlopen(request, timeout=timeout)
+            else:
+                response = urllib2.urlopen(request)
+        except urllib2.HTTPError, error:
+            raise HTTPHandlerError(error.filename, error.code, error.msg, dict(error.headers), error.read())
+        except urllib2.URLError, error:
+            if hasattr(error.reason, 'args') and isinstance(error.reason.args, tuple):
+                raise HTTPHandlerError(httpcode=error.reason.args[0], httpmsg=error.reason.args[1])
+            else:
+                raise HTTPHandlerError(httpmsg='urllib2.URLError: %s' % (error.reason))
+        except httplib.BadStatusLine, error:
+            raise HTTPHandlerError(httpmsg='httplib.BadStatusLine: %s' % (error.line))
+        return response.read()
+
 class Client(object):
     """
-    This is it. This class implements the json-RPC protocol to communicate with Transmission.
+    This is it. This class implements the JSON-RPC protocol to communicate with Transmission.
+
+    Torrent ids
+    -----------
+
+    Many functions in Client takes torrent id. A torrent id can either be id or
+    hashString. When suppling multiple id's it is possible to use a list mixed
+    with both id and hashString.
+
+    Timeouts
+    --------
+
+    Since most methods results in HTTP requests against Transmission, it is
+    possible to provide a argument called ``timeout``. Timeout is only effective
+    when using Python 2.6 or later and the default timeout is 30 seconds.
     """
 
-    def __init__(self, address='localhost', port=DEFAULT_PORT, user=None, password=None):
-        base_url = 'http://' + address + ':' + str(port)
-        self.url = base_url + '/transmission/rpc'
+    def __init__(self, address='localhost', port=DEFAULT_PORT, user=None, password=None, http_handler=None, timeout=None):
+        if isinstance(timeout, (int, long, float)):
+            self._query_timeout = float(timeout)
+        else:
+            self._query_timeout = DEFAULT_TIMEOUT
+        urlo = urlparse.urlparse(address)
+        if urlo.scheme == '':
+            base_url = 'http://' + address + ':' + str(port)
+            self.url = base_url + '/transmission/rpc'
+        else:
+            if urlo.port:
+                self.url = urlo.scheme + '://' + urlo.hostname + ':' + str(urlo.port) + urlo.path
+            else:
+                self.url = urlo.scheme + '://' + urlo.hostname + urlo.path
+            logger.info('Using custom URL "' + self.url + '".')
+            if urlo.username and urlo.password:
+                user = urlo.username
+                password = urlo.password
+            elif urlo.username or urlo.password:
+                logger.warning('Either user or password missing, not using authentication.')
+        if http_handler == None:
+            self.http_handler = DefaultHTTPHandler()
+        else:
+            if hasattr(http_handler, 'set_authentication') and hasattr(http_handler, 'request'):
+                self.http_handler = http_handler
+            else:
+                raise ValueError('Invalid HTTP handler.')
         if user and password:
-            password_manager = urllib2.HTTPPasswordMgrWithDefaultRealm()
-            password_manager.add_password(realm=None, uri=self.url, user=user, passwd=password)
-            opener = urllib2.build_opener(
-                urllib2.HTTPBasicAuthHandler(password_manager)
-                , urllib2.HTTPDigestAuthHandler(password_manager)
-                )
-            urllib2.install_opener(opener)
+            self.http_handler.set_authentication(self.url, user, password)
         elif user or password:
             logger.warning('Either user or password missing, not using authentication.')
         self._sequence = 0
         self.session = Session()
-        self.sessionid = 0
+        self.session_id = 0
         self.protocol_version = None
         self.get_session()
         self.torrent_get_arguments = get_arguments('torrent-get'
                                                    , self.rpc_version)
 
-    def _debug_request(self, request):
-        logger.debug(
-            json.dumps(
-                {
-                    'request': {
-                        'url': request.get_full_url(),
-                        'request-headers': dict(request.header_items()),
-                        'request-data': json.loads(request.data),
-                    }
-                },
-                indent=2
-            )
-        )
+    def get_timeout(self):
+        return self._query_timeout
+    
+    def set_timeout(self, value):
+        self._query_timeout = float(value)
+    
+    def del_timeout(self):
+        self._query_timeout = DEFAULT_TIMEOUT
+    
+    timeout = property(get_timeout, set_timeout, del_timeout, doc="Query timeout.")
 
-    def _debug_response(self, response, response_data):
+    def _debug_httperror(self, error):
         try:
-            response_data = json.loads(response_data)
-        except:
-            pass
+            data = json.loads(error.data)
+        except ValueError:
+            data = error.data
         logger.debug(
             json.dumps(
                 {
                     'response': {
-                        'url': response.url,
-                        'code': response.code,
-                        'msg': response.msg,
-                        'headers': dict(response.headers),
-                        'data': response_data,
+                        'url': error.url,
+                        'code': error.code,
+                        'msg': error.message,
+                        'headers': error.headers,
+                        'data': data,
                     }
                 },
                 indent=2
             )
         )
 
-    def _http_query(self, query, timeout=DEFAULT_TIMEOUT):
-        headers = {'X-Transmission-Session-Id': self.sessionid}
-        request = urllib2.Request(self.url, query, headers)
+    def _http_query(self, query, timeout=None):
+        headers = {'x-transmission-session-id': self.session_id}
         request_count = 0
+        if timeout == None:
+            timeout = self._query_timeout
         while True:
-            error_data = ""
+            logger.debug(json.dumps({'url': self.url, 'headers': headers, 'query': query, 'timeout': timeout}, indent=2))
             try:
-                self._debug_request(request)
-                socket.setdefaulttimeout(timeout) # 30 seconds
-                if (sys.version_info[0] == 2 and sys.version_info[1] > 5) or sys.version_info[0] > 2:
-                    response = urllib2.urlopen(request, timeout=timeout)
-                else:
-                    response = urllib2.urlopen(request)
+                result = self.http_handler.request(self.url, query, headers, timeout)
                 break
-            except urllib2.HTTPError, error:
-                error_data = error.read()
+            except HTTPHandlerError, error:
                 if error.code == 409:
                     logger.info('Server responded with 409, trying to set session-id.')
                     if request_count > 1:
                         raise TransmissionError('Session ID negotiation failed.', error)
-                    if 'X-Transmission-Session-Id' in error.headers:
-                        self.sessionid = error.headers['X-Transmission-Session-Id']
-                        request.add_header('X-Transmission-Session-Id', self.sessionid)
+                    if 'x-transmission-session-id' in error.headers:
+                        self.session_id = error.headers['x-transmission-session-id']
+                        headers = {'x-transmission-session-id': self.session_id}
                     else:
                         raise TransmissionError('Unknown conflict.', error)
                 else:
-                    raise TransmissionError('Request failed with %s.' % (str(error)), error)
-            except urllib2.URLError, error:
-                raise TransmissionError('Request failed.', error)
-            except httplib.BadStatusLine, error:
-                if (request_count > 1):
-                    raise TransmissionError('Failed to request %s "%s".' % (self.url, query), error)
-            finally:
-                if error_data:
-                    self._debug_response(error, error_data)
+                    raise TransmissionError('Request failed.', error)
+                self._debug_httperror(error)
             request_count = request_count + 1
-        result = response.read()
-        self._debug_response(response, result)
         return result
 
-    def _request(self, method, arguments={}, ids=[], require_ids=False, timeout=DEFAULT_TIMEOUT):
+    def _request(self, method, arguments={}, ids=[], require_ids=False, timeout=None):
         """Send json-rpc request to Transmission using http POST"""
 
         if not isinstance(method, (str, unicode)):
@@ -302,7 +354,6 @@ class Client(object):
 
         query = json.dumps({'tag': self._sequence, 'method': method
                             , 'arguments': arguments})
-        logger.info(query)
         self._sequence += 1
         start = time.time()
         http_data = self._http_query(query, timeout)
@@ -317,7 +368,7 @@ class Client(object):
             logger.error('HTTP data: \"%s\"' % (http_data))
             raise
 
-        logger.info(json.dumps(data, indent=2))
+        logger.debug(json.dumps(data, indent=2))
         if 'result' in data:
             if data['result'] != 'success':
                 raise TransmissionError('Query failed with result \"%s\".' % (data['result']))
@@ -369,7 +420,7 @@ class Client(object):
                     try:
                         int(item, 16)
                         addition = [item]
-                    except:
+                    except ValueError:
                         pass
                 if not addition:
                     # handle index ranges i.e. 5:10
@@ -379,7 +430,7 @@ class Client(object):
                             idx_from = int(match.group(1))
                             idx_to = int(match.group(2))
                             addition = range(idx_from, idx_to + 1)
-                        except:
+                        except ValueError:
                             pass
                 if not addition:
                     raise ValueError(u'Invalid torrent id, \"%s\"' % item)
@@ -409,23 +460,26 @@ class Client(object):
         if self.rpc_version < version:
             logger.warning('Using feature not supported by server. RPC version for server %d, feature introduced in %d.' % (self.rpc_version, version))
 
-    def add(self, data, timeout=DEFAULT_TIMEOUT, **kwargs):
+    def add(self, data, timeout=None, **kwargs):
         """
         Add torrent to transfers list. Takes a base64 encoded .torrent file in data.
         Additional arguments are:
 
-            * `metainfo`, string, alternate way to pass base64 encoded torrent data
-            * `filename`, path or url, provide torrent data as a file path or URL.
-            * `paused`, boolean, Whether to pause the transfer on add.
-            * `download_dir`, path, The directory where the downloaded
-              contents will be saved in.
-            * `peer_limit`, number, Limits the number of peers for this
-              transfer.
-            * `files_unwanted`,
-            * `files_wanted`,
-            * `priority_high`,
-            * `priority_low`,
-            * `priority_normal`,
+        ===================== ==== =============================================================
+        Argument              RPC  Description                                                  
+        ===================== ==== =============================================================
+        ``bandwidthPriority`` 8 -  Priority for this transfer.                                  
+        ``download_dir``      1 -  The directory where the downloaded contents will be saved in.
+        ``filename``          1 -  A filepath or URL to a torrent file or a magnet link.        
+        ``files_unwanted``    1 -  A list of file id's that shouldn't be downloaded.            
+        ``files_wanted``      1 -  A list of file id's that should be downloaded.               
+        ``metainfo``          1 -  The content of a torrent file, base64 encoded.               
+        ``paused``            1 -  If True, does not start the transfer when added.             
+        ``peer_limit``        1 -  Maximum number of peers allowed.                             
+        ``priority_high``     1 -  A list of file id's that should have high priority.          
+        ``priority_low``      1 -  A list of file id's that should have low priority.           
+        ``priority_normal``   1 -  A list of file id's that should have normal priority.        
+        ===================== ==== =============================================================
         """
         args = {}
         if data:
@@ -444,16 +498,19 @@ class Client(object):
         Add torrent to transfers list. Takes a url to a .torrent file.
         Additional arguments are:
 
-            * `paused`, boolean, Whether to pause the transfer on add.
-            * `download_dir`, path, The directory where the downloaded
-              contents will be saved in.
-            * `peer_limit`, number, Limits the number of peers for this
-              transfer.
-            * `files_unwanted`,
-            * `files_wanted`,
-            * `priority_high`,
-            * `priority_low`,
-            * `priority_normal`,
+        ===================== ==== =============================================================
+        Argument              RPC  Description                                                  
+        ===================== ==== =============================================================
+        ``bandwidthPriority`` 8 -  Priority for this transfer.                                  
+        ``download_dir``      1 -  The directory where the downloaded contents will be saved in.
+        ``files_unwanted``    1 -  A list of file id's that shouldn't be downloaded.            
+        ``files_wanted``      1 -  A list of file id's that should be downloaded.               
+        ``paused``            1 -  If True, does not start the transfer when added.             
+        ``peer_limit``        1 -  Maximum number of peers allowed.                             
+        ``priority_high``     1 -  A list of file id's that should have high priority.          
+        ``priority_low``      1 -  A list of file id's that should have low priority.           
+        ``priority_normal``   1 -  A list of file id's that should have normal priority.        
+        ===================== ==== =============================================================
         """
         torrent_file = None
         if os.path.exists(torrent_url):
@@ -461,8 +518,14 @@ class Client(object):
         else:
             try:
                 torrent_file = urllib2.urlopen(torrent_url)
-            except:
-                torrent_file = None
+            except urllib2.HTTPError:
+                pass
+            except urllib2.URLError:
+                pass
+            except httplib.BadStatusLine:
+                pass
+            except ValueError:
+                pass
 
         if not torrent_file:
             raise TransmissionError('File does not exist.')
@@ -476,16 +539,19 @@ class Client(object):
         all uri's supported by Transmissions torrent-add 'filename'
         argument. Additional arguments are:
 
-            * `paused`, boolean, Whether to pause the transfer on add.
-            * `download_dir`, path, The directory where the downloaded
-              contents will be saved in.
-            * `peer_limit`, number, Limits the number of peers for this
-              transfer.
-            * `files_unwanted`,
-            * `files_wanted`,
-            * `priority_high`,
-            * `priority_low`,
-            * `priority_normal`,
+        ===================== ==== =============================================================
+        Argument              RPC  Description                                                  
+        ===================== ==== =============================================================
+        ``bandwidthPriority`` 8 -  Priority for this transfer.                                  
+        ``download_dir``      1 -  The directory where the downloaded contents will be saved in.
+        ``files_unwanted``    1 -  A list of file id's that shouldn't be downloaded.            
+        ``files_wanted``      1 -  A list of file id's that should be downloaded.               
+        ``paused``            1 -  If True, does not start the transfer when added.             
+        ``peer_limit``        1 -  Maximum number of peers allowed.                             
+        ``priority_high``     1 -  A list of file id's that should have high priority.          
+        ``priority_low``      1 -  A list of file id's that should have low priority.           
+        ``priority_normal``   1 -  A list of file id's that should have normal priority.        
+        ===================== ==== =============================================================
         """
         if uri == None:
             raise ValueError('add_uri requires a URI.')
@@ -501,7 +567,7 @@ class Client(object):
         else:
             return self.add(None, filename=uri, **kwargs)
     
-    def remove(self, ids, delete_data=False, timeout=DEFAULT_TIMEOUT):
+    def remove(self, ids, delete_data=False, timeout=None):
         """
         remove torrent(s) with provided id(s). Local data is removed if
         delete_data is True, otherwise not.
@@ -510,30 +576,30 @@ class Client(object):
         self._request('torrent-remove',
                     {'delete-local-data':rpc_bool(delete_data)}, ids, True, timeout=timeout)
 
-    def start(self, ids, timeout=DEFAULT_TIMEOUT):
+    def start(self, ids, timeout=None):
         """start torrent(s) with provided id(s)"""
         self._request('torrent-start', {}, ids, True, timeout=timeout)
 
-    def stop(self, ids, timeout=DEFAULT_TIMEOUT):
+    def stop(self, ids, timeout=None):
         """stop torrent(s) with provided id(s)"""
         self._request('torrent-stop', {}, ids, True, timeout=timeout)
 
-    def verify(self, ids, timeout=DEFAULT_TIMEOUT):
+    def verify(self, ids, timeout=None):
         """verify torrent(s) with provided id(s)"""
         self._request('torrent-verify', {}, ids, True, timeout=timeout)
 
-    def reannounce(self, ids, timeout=DEFAULT_TIMEOUT):
-        """reannounce torrent(s) with provided id(s)"""
+    def reannounce(self, ids, timeout=None):
+        """Reannounce torrent(s) with provided id(s)"""
         self._rpc_version_warning(5)
         self._request('torrent-reannounce', {}, ids, True, timeout=timeout)
 
-    def info(self, ids=[], arguments={}, timeout=DEFAULT_TIMEOUT):
+    def info(self, ids=[], arguments={}, timeout=None):
         """Get detailed information for torrent(s) with provided id(s)."""
         if not arguments:
             arguments = self.torrent_get_arguments
         return self._request('torrent-get', {'fields': arguments}, ids, timeout=timeout)
 
-    def get_files(self, ids=[], timeout=DEFAULT_TIMEOUT):
+    def get_files(self, ids=[], timeout=None):
         """
         Get list of files for provided torrent id(s).
         This function returns a dictonary for each requested torrent id holding
@@ -542,11 +608,11 @@ class Client(object):
         fields = ['id', 'name', 'hashString', 'files', 'priorities', 'wanted']
         request_result = self._request('torrent-get', {'fields': fields}, ids, timeout=timeout)
         result = {}
-        for id, torrent in request_result.iteritems():
-            result[id] = torrent.files()
+        for tid, torrent in request_result.iteritems():
+            result[tid] = torrent.files()
         return result
 
-    def set_files(self, items, timeout=DEFAULT_TIMEOUT):
+    def set_files(self, items, timeout=None):
         """
         Set file properties. Takes a dictonary with similar contents as the
         result of get_files.
@@ -561,19 +627,19 @@ class Client(object):
             priority_high = []
             priority_normal = []
             priority_low = []
-            for fid, file in files.iteritems():
-                if not isinstance(file, dict):
+            for fid, file_desc in files.iteritems():
+                if not isinstance(file_desc, dict):
                     continue
-                if 'selected' in file and file['selected']:
+                if 'selected' in file_desc and file_desc['selected']:
                     wanted.append(fid)
                 else:
                     unwanted.append(fid)
-                if 'priority' in file:
-                    if file['priority'] == 'high':
+                if 'priority' in file_desc:
+                    if file_desc['priority'] == 'high':
                         priority_high.append(fid)
-                    elif file['priority'] == 'normal':
+                    elif file_desc['priority'] == 'normal':
                         priority_normal.append(fid)
-                    elif file['priority'] == 'low':
+                    elif file_desc['priority'] == 'low':
                         priority_low.append(fid)
             self.change([tid], files_wanted = wanted
                         , files_unwanted = unwanted
@@ -581,14 +647,14 @@ class Client(object):
                         , priority_normal = priority_normal
                         , priority_low = priority_low, timeout=timeout)
 
-    def list(self, timeout=DEFAULT_TIMEOUT):
+    def list(self, timeout=None):
         """list all torrents"""
         fields = ['id', 'hashString', 'name', 'sizeWhenDone', 'leftUntilDone'
             , 'eta', 'status', 'rateUpload', 'rateDownload', 'uploadedEver'
             , 'downloadedEver']
         return self._request('torrent-get', {'fields': fields}, timeout=timeout)
 
-    def change(self, ids, timeout=DEFAULT_TIMEOUT, **kwargs):
+    def change(self, ids, timeout=None, **kwargs):
         """
         Change torrent parameters. This is the list of parameters that.
         """
@@ -604,24 +670,24 @@ class Client(object):
         else:
             ValueError("No arguments to set")
     
-    def move(self, ids, location, timeout=DEFAULT_TIMEOUT):
+    def move(self, ids, location, timeout=None):
         """Move torrent data to the new location."""
         self._rpc_version_warning(6)
         args = {'location': location, 'move': True}
-        self._request('torrent-set-location', args, ids, True, timeout=timeout);
+        self._request('torrent-set-location', args, ids, True, timeout=timeout)
     
-    def locate(self, ids, location, timeout=DEFAULT_TIMEOUT):
+    def locate(self, ids, location, timeout=None):
         """Locate torrent data at the location."""
         self._rpc_version_warning(6)
         args = {'location': location, 'move': False}
-        self._request('torrent-set-location', args, ids, True, timeout=timeout);
+        self._request('torrent-set-location', args, ids, True, timeout=timeout)
     
-    def get_session(self, timeout=DEFAULT_TIMEOUT):
+    def get_session(self, timeout=None):
         """Get session parameters"""
         self._request('session-get', timeout=timeout)
         return self.session
 
-    def set_session(self, timeout=DEFAULT_TIMEOUT, **kwargs):
+    def set_session(self, timeout=None, **kwargs):
         """Set session parameters"""
         args = {}
         for key, value in kwargs.iteritems():
@@ -634,7 +700,7 @@ class Client(object):
         if len(args) > 0:
             self._request('session-set', args, timeout=timeout)
 
-    def blocklist_update(self, timeout=DEFAULT_TIMEOUT):
+    def blocklist_update(self, timeout=None):
         """Update block list. Returns the size of the block list."""
         self._rpc_version_warning(5)
         result = self._request('blocklist-update', timeout=timeout)
@@ -642,7 +708,7 @@ class Client(object):
             return result['blocklist-size']
         return None
 
-    def port_test(self, timeout=DEFAULT_TIMEOUT):
+    def port_test(self, timeout=None):
         """
         Tests to see if your incoming peer port is accessible from the
         outside world.
@@ -653,7 +719,7 @@ class Client(object):
             return result['port-is-open']
         return None
 
-    def session_stats(self, timeout=DEFAULT_TIMEOUT):
+    def session_stats(self, timeout=None):
         """Get session statistics"""
         self._request('session-stats', timeout=timeout)
         return self.session
